@@ -1,8 +1,8 @@
-use std::io::{copy, Write};
+use std::io::{copy, stdin, stdout, Write};
 
 fn main() -> anyhow::Result<()> {
-    let mut reader = std::io::stdin();
-    let mut writer = std::io::stdout();
+    let mut reader = stdin();
+    let mut writer = stdout();
     copy(&mut reader, &mut writer)?;
     writer.flush()?;
     Ok(())
