@@ -1,23 +1,28 @@
-//! Input for `TextReader` and the reader half of `TextDuplexer`.
+//! Shared implementation for `TextReader` and the reader half of
+//! `TextDuplexer`.
 
-use crate::{
-    replace_selected::ReplaceSelected,
-    text_utils::{is_end_ok, is_start_ok},
+use crate::{TextDuplexer, TextReader, TextSubstr};
+use basic_text_internals::{
+    is_basic_text_end, is_basic_text_start, replace,
     unicode::{
-        BEL, BOM, CAN, CGJ, DEL, ESC, FF, MAX_UTF8_SIZE, NEL, NORMALIZATION_BUFFER_SIZE, REPL,
+        BEL, BOM, CAN, CGJ, DEL, ESC, LS, MAX_UTF8_SIZE, NEL, NORMALIZATION_BUFFER_SIZE, PS, REPL,
     },
-    TextDuplexer, TextReader, TextStr,
 };
 use layered_io::{default_read, HalfDuplexLayered, Status, WriteLayered};
 use std::{
     cmp::max,
     collections::{vec_deque, VecDeque},
-    io::{self, copy, repeat, Cursor, Read},
-    mem, str,
+    io,
+    mem::take,
+    str,
+};
+use unicode_normalization::{
+    is_nfc_stream_safe_quick, IsNormalized, Recompositions, Replacements, StreamSafe,
+    UnicodeNormalization,
 };
-use unicode_normalization::{Recompositions, Replacements, StreamSafe, UnicodeNormalization};
 use utf8_io::{ReadStrLayered, WriteStr};
 
+/// Abstract over `TextReader` and the reader half of `TextDuplexer`.
 pub(crate) trait TextReaderInternals<Inner: ReadStrLayered>: ReadStrLayered {
     fn impl_(&mut self) -> &mut TextInput;
     fn inner(&self) -> &Inner;
@@ -74,8 +79,11 @@ pub(crate) struct TextInput {
     queue: VecDeque<char>,
 
     /// An iterator over the chars in `self.queue`.
-    queue_iter:
-        Recompositions<StreamSafe<ReplaceSelected<Replacements<vec_deque::IntoIter<char>>>>>,
+    ssnfc_iter: Recompositions<StreamSafe<Replacements<vec_deque::IntoIter<char>>>>,
+
+    /// The number of characters in the queue which are already verified to be
+    /// Stream-Safe NFC and can skip normalization.
+    quick: usize,
 
     /// When we can't fit all the data from an underlying read in our buffer,
     /// we buffer it up. Remember the status value so we can replay that too.
@@ -85,9 +93,15 @@ pub(crate) struct TextInput {
     /// normalization-form starter.
     expect_starter: bool,
 
-    /// For emitting BOM at the start of a stream.
+    /// For ignoring BOM at the start of a stream.
     at_start: bool,
 
+    /// NEL compatibility mode.
+    nel_compatibility: bool,
+
+    /// LSPS compatibility mode.
+    lsps_compatibility: bool,
+
     /// Control-code and escape-sequence state machine.
     state: State,
 }
@@ -100,18 +114,37 @@ impl TextInput {
         Self {
             raw_string: String::new(),
             queue,
-            queue_iter: ReplaceSelected::new(
-                VecDeque::<char>::new().into_iter().cjk_compat_variants(),
-            )
-            .stream_safe()
-            .nfc(),
+            ssnfc_iter: VecDeque::<char>::new()
+                .into_iter()
+                .cjk_compat_variants()
+                .stream_safe()
+                .nfc(),
+            quick: 0,
             pending_status: Status::active(),
             expect_starter: true,
             at_start: true,
+            nel_compatibility: false,
+            lsps_compatibility: false,
             state: State::Ground(true),
         }
     }
 
+    /// Construct a new instance of `TextInput` in NEL compatibility mode.
+    #[inline]
+    pub(crate) fn with_nel_compatibility() -> Self {
+        let mut result = Self::new();
+        result.nel_compatibility = true;
+        result
+    }
+
+    /// Construct a new instance of `TextInput` in LSPS compatibility mode.
+    #[inline]
+    pub(crate) fn with_lsps_compatibility() -> Self {
+        let mut result = Self::new();
+        result.lsps_compatibility = true;
+        result
+    }
+
     /// Like `read_with_status` but produces the result in a `str`. Be sure to
     /// check the `size` field of the return value to see how many bytes were
     /// written.
@@ -162,124 +195,127 @@ impl TextInput {
         internals.read_exact_using_status(unsafe { buf.as_bytes_mut() })
     }
 
-    /// Like `read_with_status` but produces the result in a `TextStr`. Be sure
-    /// to check the `size` field of the return value to see how many bytes
-    /// were written.
     #[inline]
-    pub(crate) fn read_text<Inner: ReadStrLayered>(
+    pub(crate) fn read_text_substr<Inner: ReadStrLayered>(
         internals: &mut impl TextReaderInternals<Inner>,
-        buf: &mut TextStr,
+        buf: &mut TextSubstr,
     ) -> io::Result<usize> {
-        // Safety: This is a UTF-8 stream so we can read directly into a
-        // `TextStr`.
+        // Safety: This is a Basic Text stream so we can read directly into a
+        // `TextSubstr`.
         internals.read(unsafe { buf.as_bytes_mut() })
     }
 
-    /// Like `read_exact` but produces the result in a `TextStr`.
     #[inline]
-    pub(crate) fn read_exact_text<Inner: ReadStrLayered>(
+    pub(crate) fn read_exact_text_substr<Inner: ReadStrLayered>(
         internals: &mut impl TextReaderInternals<Inner>,
-        buf: &mut TextStr,
+        buf: &mut TextSubstr,
     ) -> io::Result<()> {
-        // Safety: This is a Text stream so we can read directly into a
-        // `TextStr`.
+        // Safety: This is a Basic Text stream so we can read directly into a
+        // `TextSubstr`.
         internals.read_exact(unsafe { buf.as_bytes_mut() })
     }
 
-    /// Like `read_with_status` but produces the result in a `TextStr`. Be sure
-    /// to check the `size` field of the return value to see how many bytes
-    /// were written.
     #[inline]
-    pub(crate) fn read_text_with_status<Inner: ReadStrLayered>(
+    pub(crate) fn read_text_substr_with_status<Inner: ReadStrLayered>(
         internals: &mut impl TextReaderInternals<Inner>,
-        buf: &mut TextStr,
+        buf: &mut TextSubstr,
     ) -> io::Result<(usize, Status)> {
-        // Safety: This is a text stream so we can read directly into a `TextStr`.
-        let (size, status) = internals.read_with_status(unsafe { buf.as_bytes_mut() })?;
-
-        // TODO
-        //debug_assert!(buf.is_char_boundary(size));
-
-        Ok((size, status))
+        // Safety: This is a Basic Text stream so we can read directly into a
+        // `TextSubstr`.
+        internals.read_with_status(unsafe { buf.as_bytes_mut() })
     }
 
-    /// Like `read_with_status` but produces the result in a `TextStr`. Be sure
-    /// to check the `size` field of the return value to see how many bytes
-    /// were written.
     #[inline]
-    pub(crate) fn read_exact_text_using_status<Inner: ReadStrLayered>(
+    pub(crate) fn read_exact_text_substr_using_status<Inner: ReadStrLayered>(
         internals: &mut impl TextReaderInternals<Inner>,
-        buf: &mut TextStr,
+        buf: &mut TextSubstr,
     ) -> io::Result<Status> {
-        // Safety: This is a text stream so we can read directly into a `str`.
+        // Safety: This is a Basic Text stream so we can read directly into a
+        // `TextSubstr`.
         internals.read_exact_using_status(unsafe { buf.as_bytes_mut() })
     }
 
     fn queue_next(&mut self) -> Option<char> {
-        match self.queue_iter.next() {
-            Some(c) => Some(c),
-            None => {
-                let index = self.queue.iter().position(|c| matches!(*c, '\n' | CGJ))?;
-                let tmp = self.queue.drain(0..=index).collect::<VecDeque<char>>();
-                self.queue_iter = ReplaceSelected::new(tmp.into_iter().cjk_compat_variants())
-                    .stream_safe()
-                    .nfc();
-                self.queue_iter.next()
+        let quick = self.quick;
+        if quick != 0 {
+            self.quick = quick - 1;
+            self.queue.pop_front()
+        } else {
+            match self.ssnfc_iter.next() {
+                Some(c) => Some(c),
+                None => {
+                    let last_boundary = self
+                        .queue
+                        .iter()
+                        .rev()
+                        .position(|c| matches!(*c, '\n' | CGJ))?;
+                    let index = self.queue.len() - last_boundary;
+                    if is_nfc_stream_safe_quick(self.queue.iter().take(index).copied())
+                        == IsNormalized::Yes
+                    {
+                        self.quick = index - 1;
+                        self.queue.pop_front()
+                    } else {
+                        let tmp = self.queue.drain(..index).collect::<VecDeque<char>>();
+                        self.ssnfc_iter = tmp.into_iter().cjk_compat_variants().stream_safe().nfc();
+                        self.ssnfc_iter.next()
+                    }
+                }
             }
         }
     }
 
     fn process_raw_string(&mut self) {
-        for c in self.raw_string.chars() {
-            let at_start = mem::take(&mut self.at_start);
+        let mut chars = self.raw_string.chars();
+
+        // If we're at the start of a stream, skip over a leading BOM.
+        if take(&mut self.at_start) && self.raw_string.starts_with(BOM) {
+            chars.next();
+        }
+
+        for c in chars {
             loop {
                 match (self.state, c) {
-                    (State::Ground(_), BOM) if at_start => (),
-                    (State::Ground(_), '\n') => {
-                        self.queue.push_back('\n');
-                        self.expect_starter = false;
-                        self.state = State::Ground(true);
-                    }
-                    (State::Ground(_), '\t') => {
-                        self.queue.push_back('\t');
-                        self.expect_starter = false;
-                        self.state = State::Ground(false);
-                    }
-                    (State::Ground(_), FF) | (State::Ground(_), NEL) => {
-                        self.queue.push_back(' ');
-                        self.expect_starter = false;
-                        self.state = State::Ground(false);
-                    }
-                    (State::Ground(_), '\r') => self.state = State::Cr,
-                    (State::Ground(_), ESC) => self.state = State::Esc,
-                    (State::Ground(_), c)
-                        if c.is_control() || matches!(c, '\u{2329}' | '\u{232a}') =>
-                    {
-                        self.queue.push_back(REPL);
-                        self.expect_starter = false;
-                        self.state = State::Ground(false);
-                    }
-                    (State::Ground(_), mut c) => {
-                        if self.expect_starter {
+                    (State::Ground(_), c) => match c {
+                        '\n' => {
+                            self.queue.push_back('\n');
                             self.expect_starter = false;
-                            if !is_start_ok(c) {
-                                c = REPL;
+                            self.state = State::Ground(true);
+                        }
+                        '\r' => self.state = State::Cr,
+                        ESC => self.state = State::Esc,
+                        mut c => {
+                            self.state = State::Ground(false);
+                            if (self.nel_compatibility && c == NEL)
+                                || (self.lsps_compatibility && matches!(c, LS | PS))
+                            {
+                                c = '\n';
+                                self.state = State::Ground(true);
+                            }
+                            let pos = self.queue.len();
+                            replace(c, &mut self.queue);
+
+                            // Prepend a CGJ if needed to guard a non-starter.
+                            if take(&mut self.expect_starter)
+                                && !self
+                                    .queue
+                                    .get(pos)
+                                    .copied()
+                                    .map(is_basic_text_start)
+                                    .unwrap_or(true)
+                            {
+                                self.queue.insert(pos, CGJ);
                             }
                         }
-                        self.queue.push_back(c);
-                        self.state = State::Ground(false);
-                    }
+                    },
 
-                    (State::Cr, '\n') => {
+                    (State::Cr, c) => {
                         self.queue.push_back('\n');
                         self.expect_starter = false;
                         self.state = State::Ground(true);
-                    }
-                    (State::Cr, _) => {
-                        self.queue.push_back('\n');
-                        self.expect_starter = false;
-                        self.state = State::Ground(true);
-                        continue;
+                        if c != '\n' {
+                            continue;
+                        }
                     }
 
                     (State::Esc, '[') => self.state = State::CsiStart,
@@ -337,13 +373,14 @@ impl TextInput {
         }
 
         let mut nread = 0;
-
         loop {
             match internals.impl_().queue_next() {
                 Some(c) => nread += c.encode_utf8(&mut buf[nread..]).len(),
                 None => break,
             }
             if buf.len() - nread < MAX_UTF8_SIZE {
+                // Write out single-byte codepoints to preserve UTF-8 validity.
+                clear_to_char_boundary(&mut buf[nread..]);
                 return Ok((nread, Status::active()));
             }
         }
@@ -351,20 +388,13 @@ impl TextInput {
             internals.impl_().pending_status = Status::active();
             internals.impl_().expect_starter = true;
 
-            // We may have overwritten part of a codepoint; overwrite the rest
-            // of the buffer.
-            // TODO: Use [`fill`] when it becomes available:
-            // https://doc.rust-lang.org/std/primitive.slice.html#method.fill
-            copy(
-                &mut repeat(b'?').take((buf.len() - nread) as u64),
-                &mut Cursor::new(&mut buf[nread..]),
-            )
-            .unwrap();
+            // Write out single-byte codepoints to preserve UTF-8 validity.
+            clear_to_char_boundary(&mut buf[nread..]);
 
             return Ok((nread, internals.impl_().pending_status));
         }
 
-        let mut raw_bytes = mem::take(&mut internals.impl_().raw_string).into_bytes();
+        let mut raw_bytes = take(&mut internals.impl_().raw_string).into_bytes();
         raw_bytes.resize(4096, 0_u8);
         let (size, status) = internals.inner_mut().read_with_status(&mut raw_bytes)?;
         raw_bytes.resize(size, 0);
@@ -391,10 +421,10 @@ impl TextInput {
                 }
             }
 
-            // If the stream ends in a non-ending char, append a REPL.
+            // If the stream ends in a non-ending char, append a CGJ.
             if let Some(last) = internals.impl_().queue.back() {
-                if !is_end_ok(*last) {
-                    internals.impl_().queue.push_back(REPL);
+                if !is_basic_text_end(*last) {
+                    internals.impl_().queue.push_back(CGJ);
                 }
             }
 
@@ -422,13 +452,8 @@ impl TextInput {
             }
         }
 
-        // We may have overwritten part of a codepoint; overwrite the rest
-        // of the buffer.
-        copy(
-            &mut repeat(b'?').take((buf.len() - nread) as u64),
-            &mut Cursor::new(&mut buf[nread..]),
-        )
-        .unwrap();
+        // Write out single-byte codepoints to preserve UTF-8 validity.
+        clear_to_char_boundary(&mut buf[nread..]);
 
         Ok((
             nread,
@@ -492,6 +517,40 @@ impl TextInput {
     }
 }
 
+#[inline]
+fn clear_to_char_boundary(buf: &mut [u8]) {
+    for b in buf {
+        if is_char_boundary(*b) {
+            break;
+        }
+        *b = b'\0';
+    }
+}
+
+#[test]
+fn test_clear_to_char_boundary() {
+    let mut buf = vec![7, 7, 0x80, 0x80, 7];
+    clear_to_char_boundary(&mut buf[2..]);
+    assert_eq!(buf, vec![7, 7, 0, 0, 7]);
+}
+
+#[inline]
+const fn is_char_boundary(b: u8) -> bool {
+    b as i8 >= -0x40
+}
+
+#[test]
+fn test_is_char_boundary() {
+    assert!(is_char_boundary(b'\0'));
+    assert!(is_char_boundary(b'A'));
+    assert!(is_char_boundary(0x7f));
+    assert!(!is_char_boundary(0x80));
+    assert!(!is_char_boundary(0xbf));
+    assert!(is_char_boundary(0xc0));
+    assert!(is_char_boundary(0xcf));
+    assert!(is_char_boundary(0xff));
+}
+
 #[derive(Clone, Copy, PartialEq, Eq)]
 enum State {
     // Default state. Boolean is true iff we just saw a '\n'.
