@@ -1,10 +1,11 @@
-//! Output for `TextWriter` and the writer half of `TextDuplexer`.
+//! Shared implementation for `TextWriter` and the writer half of
+//! `TextDuplexer`.
 
-use crate::{
-    categorize::Categorize,
-    text_utils::{is_end_ok, is_start_ok},
+use crate::{TextDuplexer, TextSubstr, TextWriter};
+use basic_text_internals::{
+    is_basic_text_end, is_basic_text_start,
     unicode::{BOM, ESC, MAX_UTF8_SIZE, SUB},
-    TextDuplexer, TextStr, TextWriter,
+    BasicTextError, Categorize,
 };
 #[cfg(can_vector)]
 use layered_io::default_is_write_vectored;
@@ -14,14 +15,14 @@ use layered_io::{default_write_vectored, HalfDuplexLayered, WriteLayered};
 use std::{
     cell::RefCell,
     io::{self, Write},
-    mem,
     mem::take,
     rc::Rc,
     str,
 };
-use unicode_normalization::UnicodeNormalization;
+use unicode_normalization::{is_nfc_stream_safe_quick, IsNormalized, UnicodeNormalization};
 use utf8_io::{ReadStrLayered, WriteStr};
 
+/// Abstract over `TextWriter` and the writer half of `TextDuplexer`.
 pub(crate) trait TextWriterInternals<Inner: WriteStr + WriteLayered>: Write {
     fn impl_(&mut self) -> &mut TextOutput;
     fn inner(&self) -> &Inner;
@@ -179,7 +180,8 @@ impl TextOutput {
         internals: &mut impl TextWriterInternals<Inner>,
         s: &str,
     ) -> io::Result<()> {
-        Self::state_machine(internals, s)?;
+        Self::state_machine(internals, s)
+            .map_err(|e| io::Error::new(io::ErrorKind::InvalidData, e))?;
 
         // Write to the underlying stream.
         Self::write_buffer(internals)
@@ -200,7 +202,8 @@ impl TextOutput {
                 impl_.buffer.push_str("\r\n");
             }
 
-            Self::state_machine(internals, slice)?;
+            Self::state_machine(internals, slice)
+                .map_err(|e| io::Error::new(io::ErrorKind::InvalidData, e))?;
         }
 
         // Write to the underlying stream.
@@ -209,7 +212,7 @@ impl TextOutput {
 
     fn normal_write_text<Inner: WriteStr + WriteLayered>(
         internals: &mut impl TextWriterInternals<Inner>,
-        s: &TextStr,
+        s: &TextSubstr,
     ) -> io::Result<()> {
         let impl_ = internals.impl_();
         let s = s.as_ref(); // TODO: Avoid doing this.
@@ -217,7 +220,7 @@ impl TextOutput {
         impl_.buffer.push_str(s);
         let ground = match s.chars().next_back() {
             None | Some('\n') => Ground::Newline,
-            Some(c) if !is_end_ok(c) => Ground::ZwjOrPrepend,
+            Some(c) if !is_basic_text_end(c) => Ground::ZwjOrPrepend,
             _ => Ground::Other,
         };
         impl_.state = State::Ground(ground);
@@ -228,7 +231,7 @@ impl TextOutput {
 
     fn crlf_write_text<Inner: WriteStr + WriteLayered>(
         internals: &mut impl TextWriterInternals<Inner>,
-        s: &TextStr,
+        s: &TextSubstr,
     ) -> io::Result<()> {
         let s: &str = s.as_ref(); // TODO: Avoid doing this.
 
@@ -246,7 +249,7 @@ impl TextOutput {
             impl_.buffer.push_str(slice);
             let ground = match slice.chars().next_back() {
                 None | Some('\n') => Ground::Newline,
-                Some(c) if !is_end_ok(c) => Ground::ZwjOrPrepend,
+                Some(c) if !is_basic_text_end(c) => Ground::ZwjOrPrepend,
                 _ => Ground::Other,
             };
             impl_.state = State::Ground(ground);
@@ -262,10 +265,10 @@ impl TextOutput {
         if internals.impl_().expect_starter {
             internals.impl_().expect_starter = false;
             if let Some(c) = internals.impl_().buffer.chars().next() {
-                if !is_start_ok(c) {
+                if !is_basic_text_start(c) {
                     Self::reset_state(internals);
                     return Err(io::Error::new(
-                        io::ErrorKind::Other,
+                        io::ErrorKind::InvalidData,
                         "write data must begin with a Unicode Normalization Form starter",
                     ));
                 }
@@ -291,68 +294,97 @@ impl TextOutput {
     fn state_machine<Inner: WriteStr + WriteLayered>(
         internals: &mut impl TextWriterInternals<Inner>,
         s: &str,
-    ) -> io::Result<()> {
+    ) -> Result<(), BasicTextError> {
         let error = Rc::new(RefCell::new(None));
-        let impl_ = internals.impl_();
-        for c in Categorize::new(s.chars(), Rc::clone(&error))
+
+        if is_nfc_stream_safe_quick(s.chars()) == IsNormalized::Yes {
+            // Fast path: Data is already Stream-Safe and NFC. Just check it
+            // for errors.
+            for c in Categorize::new(s.chars(), Rc::clone(&error)) {
+                Self::state_machine_char(internals, c, &error)?;
+            }
+        } else {
+            Self::state_machine_slow_path(internals, s, &error)?;
+        }
+
+        Ok(())
+    }
+
+    #[cold]
+    fn state_machine_slow_path<Inner: WriteStr + WriteLayered>(
+        internals: &mut impl TextWriterInternals<Inner>,
+        s: &str,
+        error: &Rc<RefCell<Option<BasicTextError>>>,
+    ) -> Result<(), BasicTextError> {
+        // Slow path: Compute Stream-Safe and NFC.
+        for c in Categorize::new(s.chars(), Rc::clone(error))
             .cjk_compat_variants()
             .stream_safe()
             .nfc()
         {
-            match (&impl_.state, c) {
-                // Recognize ANSI-style color escape sequences.
-                (State::Ground(_), ESC) if impl_.ansi_color => {
-                    impl_.state = State::Esc;
-                    impl_.escape_sequence.clear();
-                    impl_.escape_sequence.push(ESC);
-                }
-                (State::Esc, '[') => {
-                    impl_.state = State::Csi;
-                    impl_.escape_sequence.push('[');
-                }
-                (State::Csi, c) if matches!(c, ' '..='?') => impl_.escape_sequence.push(c),
-                (State::Csi, 'm') => {
-                    impl_.escape_sequence.push('m');
-                    impl_.buffer.push_str(&impl_.escape_sequence);
-                    impl_.state = State::Ground(Ground::Other);
-                }
+            Self::state_machine_char(internals, c, error)?;
+        }
 
-                (State::Ground(_), '\n') => {
-                    impl_.state = State::Ground(Ground::Newline);
-                    impl_.buffer.push(c);
-                }
+        Ok(())
+    }
 
-                (State::Ground(_), SUB) => {
-                    // SUB indicates an error sent through the NFC iterator
-                    // chain, and the Rc<RefCell<Option<io::Error>>> holds the
-                    // actual error.
-                    Self::reset_state(internals);
-                    return Err(mem::take(&mut *error.borrow_mut()).unwrap());
-                }
+    fn state_machine_char<Inner: WriteStr + WriteLayered>(
+        internals: &mut impl TextWriterInternals<Inner>,
+        c: char,
+        error: &Rc<RefCell<Option<BasicTextError>>>,
+    ) -> Result<(), BasicTextError> {
+        let impl_ = internals.impl_();
+        match (&impl_.state, c) {
+            // Recognize ANSI-style color escape sequences.
+            (State::Ground(_), ESC) if impl_.ansi_color => {
+                impl_.state = State::Esc;
+                impl_.escape_sequence.clear();
+                impl_.escape_sequence.push(ESC);
+            }
+            (State::Esc, '[') => {
+                impl_.state = State::Csi;
+                impl_.escape_sequence.push('[');
+            }
+            (State::Csi, c) if matches!(c, ' '..='?') => impl_.escape_sequence.push(c),
+            (State::Csi, 'm') => {
+                impl_.escape_sequence.push('m');
+                impl_.buffer.push_str(&impl_.escape_sequence);
+                impl_.state = State::Ground(Ground::Other);
+            }
 
-                (State::Ground(_), ESC) => {
-                    impl_.state = State::Esc;
-                }
+            (State::Ground(_), '\n') => {
+                impl_.state = State::Ground(Ground::Newline);
+                impl_.buffer.push(c);
+            }
 
-                // Common case: in ground state and reading a normal char.
-                (State::Ground(_), c) => {
-                    let ground = if !is_end_ok(c) {
-                        Ground::ZwjOrPrepend
-                    } else {
-                        Ground::Other
-                    };
-                    impl_.state = State::Ground(ground);
-                    impl_.buffer.push(c);
-                }
+            (State::Ground(_), SUB) => {
+                // SUB indicates an error sent through the NFC iterator
+                // chain, and the Rc<RefCell<Option<BasicTextError>>> holds the
+                // actual error.
+                Self::reset_state(internals);
+                return Err(take(&mut *error.borrow_mut()).unwrap());
+            }
 
-                // Escape sequence not recognized.
-                (State::Esc, c) | (State::Csi, c) => {
-                    Self::reset_state(internals);
-                    return Err(io::Error::new(
-                        io::ErrorKind::Other,
-                        format!("unrecognized escape sequence, ending in {:?}", c),
-                    ));
-                }
+            (State::Ground(_), ESC) => {
+                Self::reset_state(internals);
+                return Err(BasicTextError::Escape);
+            }
+
+            // Common case: in ground state and reading a normal char.
+            (State::Ground(_), c) => {
+                let ground = if !is_basic_text_end(c) {
+                    Ground::ZwjOrPrepend
+                } else {
+                    Ground::Other
+                };
+                impl_.state = State::Ground(ground);
+                impl_.buffer.push(c);
+            }
+
+            // Escape sequence not recognized.
+            (State::Esc, _) | (State::Csi, _) => {
+                Self::reset_state(internals);
+                return Err(BasicTextError::UnrecognizedEscape);
             }
         }
         Ok(())
@@ -366,21 +398,21 @@ impl TextOutput {
             State::Ground(Ground::ZwjOrPrepend) => {
                 Self::reset_state(internals);
                 Err(io::Error::new(
-                    io::ErrorKind::Other,
+                    io::ErrorKind::InvalidData,
                     "output text stream closed after a ZWJ or Prepend",
                 ))
             }
             State::Ground(Ground::Other) => {
                 Self::reset_state(internals);
                 Err(io::Error::new(
-                    io::ErrorKind::Other,
+                    io::ErrorKind::InvalidData,
                     "output text stream must end with newline",
                 ))
             }
             State::Esc | State::Csi => {
                 Self::reset_state(internals);
                 Err(io::Error::new(
-                    io::ErrorKind::Other,
+                    io::ErrorKind::InvalidData,
                     "incomplete escape sequence at end of output text stream",
                 ))
             }
@@ -409,9 +441,9 @@ impl TextOutput {
         internals.inner().suggested_buffer_size()
     }
 
-    pub(crate) fn write_text<Inner: WriteStr + WriteLayered>(
+    pub(crate) fn write_text_substr<Inner: WriteStr + WriteLayered>(
         internals: &mut impl TextWriterInternals<Inner>,
-        s: &TextStr,
+        s: &TextSubstr,
     ) -> io::Result<()> {
         if internals.impl_().crlf_compatibility {
             Self::crlf_write_text(internals, s)
@@ -445,7 +477,7 @@ impl TextOutput {
             .map(|()| error.valid_up_to()),
             Err(error) => {
                 Self::reset_state(internals);
-                Err(io::Error::new(io::ErrorKind::Other, error))
+                Err(io::Error::new(io::ErrorKind::InvalidData, error))
             }
         }
     }
@@ -458,14 +490,14 @@ impl TextOutput {
             State::Ground(Ground::ZwjOrPrepend) => {
                 Self::reset_state(internals);
                 return Err(io::Error::new(
-                    io::ErrorKind::Other,
+                    io::ErrorKind::InvalidData,
                     "output text stream flushed after a ZWJ or Prepend",
                 ));
             }
             State::Ground(_) => (),
             State::Esc | State::Csi => {
                 return Err(io::Error::new(
-                    io::ErrorKind::Other,
+                    io::ErrorKind::InvalidData,
                     "output text stream flushed while an escape sequence was in progress",
                 ))
             }
